#include "trf_tarr.hpp"

class Expr{
private:

  void Expr::create_chunk(){
    ts::TVar< TArr<Term> > my_chunk;
    chunk = &my_chunk;
  }

public:
  ts::TPtr< TArr<Term> > chunk;

  ///-------------------------------------------                                                                            
  /// Constructors                                                                                                          
  ///-------------------------------------------                                                                            

  /// Create an empty expression                                                                                            
  Expr::Expr () {create_chunk(); chunk->alloc(0);}

  /// Create an expression with one term with integer value                                                                 
  /// Temporary?
  Expr::Expr (int num){
    //    printf("Construct expr\n");
    
    create_chunk();

    Term *terms;

    Term aa(num);

    terms = chunk->alloc(1);
    terms[0] = aa;
  }

  /// Create an subexpression from index
  Expr::Expr (Expr exp, uint index){
    printf("Construct sub_expr\n");
    
    assert (exp.get_len() >= index);
    uint len = exp.get_len() - index;

    create_chunk();
    
    Term *terms;
    terms = chunk->alloc(len);
    for (uint i=0; i<len; i++){
      terms[i] = exp.chunk->getData()[i + index];
    }
  }

  /// Create an subexpression from index
  Expr::Expr (Expr exp, uint index, uint length){
    printf("Construct sub_expr2\n");

    assert (exp.get_len() >= index + length);

    create_chunk();

    Term *terms;
    terms = chunk->alloc(length);
    for (uint i=0; i<length; i++){
      terms[i] = exp.chunk->getData()[i + index];
    }
  }

  /// Copy constructor                                                                                                      
  Expr::Expr (const Expr& exp) : chunk(exp.chunk) {}

  /// Create an expression with one term that is an expression                                                              
  Expr (ts::TPtr<Expr> expr){
    Term trm(expr);

    create_chunk();
    Term *terms;
    terms = chunk->alloc(1);
    terms[0] = trm;
  }

  ///-------------------------------------------
  /// Methods
  ///-------------------------------------------

  /// Get expression length (in terms)
  inline uint get_len(){
    return chunk->getSize();
  }

  /// Return term at specified index
  inline Term elementAt(uint i){
    return chunk->getData()[i];
  }

  /// Check whether a term at specified index is a symbol
  inline bool symbol_at(uint index){
    return (chunk->getData()[index].type != typeExpr);
  }

  /// Check whether an expression is flat
  inline bool is_flat(){
    uint len = get_len();
    
    for (uint i; i<len; i++){
      if (chunk->getData()[i].type == typeExpr) return false;
    }
    return true;
  }

  /// Check whether an expression is empty
  inline bool is_empty() { return (this->get_len() == 0);}

  /// Print expression
  void Expr::print(FILE *fd){
    uint len = this->get_len();

    Term cur_trm;
    for (uint i=0; i<len; i++){
      cur_trm = chunk->getData()[i];

      switch (cur_trm.type){
      case typeInt :
        fprintf(fd, "%d", cur_trm.getInt());
        break;
      case typeChar:
        fprintf(fd, "%c", cur_trm.getChar());
        break;
      case typeFloat:
        fprintf(fd, "%f", cur_trm.getFloat());
        break;
      case typeExpr:
        fprintf(fd, "(");
        cur_trm.getExpr()->print(fd);
        fprintf(fd, ")");
        break;
      case typeUninit:
        fprintf(fd, "[empty]");
        break;
      }
    }
  }
  /// PrintLN expression
  void Expr::println(FILE *fd){
    this->print(fd);
    fprintf(fd, "\n");
  }

  /// Print expression (for testing purposes)
  void Expr::my_print(){
    uint len = this->get_len();
    printf("Expr: size = %d  val = ", len);

    Term cur_trm;
    for (uint i=0; i<len; i++){
      cur_trm = chunk->getData()[i];

      switch (cur_trm.type){
      case typeInt :
        printf("[%d]", cur_trm.getInt());
        break;
      case typeChar:
        printf("[%c]", cur_trm.getChar());
        break;
      case typeFloat:
        printf("[%f]", cur_trm.getFloat());
        break;
      case typeExpr:
        printf("\033[%um\033[%um", 1, 36);
        printf("(");
        printf("\033[%um",0);
	cur_trm.getExpr()->my_print();
        printf("\033[%um\033[%um", 1, 36);
        printf(")");
        printf("\033[%um",0);
        break;
      case typeUninit:
        printf("[uninit]");
        break;
      }
    }
  }

  /// Dump expression contents (recursive)                                                                                  
  void dump(){
    this->my_print();
    printf("\n");
  }

  ///-------------------------------------------
  /// Static methods
  ///-------------------------------------------

  /// Operator enclosing an expression into parentheses                                                                     
  static Expr parentheses (ts::TPtr<Expr> in) {
    Expr res(in);
    return res;
  };

  /// static-method for expression concatenation
  static Expr concatenate (Expr exp1, Expr exp2){
    Expr res;

    Term *terms;
    terms = res.chunk->alloc(exp1.get_len() + exp2.get_len());

    uint gi=0;
    for (uint i=0; i<exp1.get_len(); i++, gi++){
      terms[gi] = exp1.elementAt(i);
    }
    for (uint i=0; i<exp2.get_len(); i++, gi++){
      terms[gi] = exp2.elementAt(i);
    }

    return res;
    
  }

};

Expr Expr_create_Integer(char *str){
  return Expr(atoi(str));
}
